import subprocess
from datetime import datetime
from pathlib import Path
from typing import Union

from ..rex import grouped_impacts_table

from dominate import document
from dominate.tags import a, div, h1, h2, img, p
from dominate.util import raw

CSS = r"""
<style>
html {
  max-width: 100ch;
  padding: 2rem;
  margin: auto;
}
body {
  font-family: Liberation Sans, Arial, sans-serif;
  background-color: #fffaf7;
  line-height: 1.5;
}
header {
  margin-bottom: 1.5rem;
}
h1 {
  margin-bottom: .5rem;
}
pre {
  white-space: pre-wrap;
}
hr {
  margin-top: 2rem;
}
.marginauto {
  margin-left: auto;
  margin-right: auto;
  width: 60%;
}
</style>
"""


def fig(v, r, s, t) -> Path:
    return f"matplotlib/VRP_reg{r}_{v}_{s}Fit.{t}"


def add_images(d, vs, r, s, t=None):
    if t is not None:
        d.add(p(t))
    for v in vs:
        d.add(a(img(src=fig(v, r, s, "png"), width="250px"), href=fig(v, r, s, "pdf")))


def index_dot_html(rex_dir: Union[str, Path]) -> None:
    rex_dir = Path(rex_dir)
    doc = document(title="tW Fit")
    with doc.head:
        raw(CSS)

    mpl_dir = rex_dir / "matplotlib"
    if not mpl_dir.exists():
        subprocess.Popen(
            f"tdub rex stacks --no-chisq --no-internal --png {rex_dir}", shell=True
        ).wait()
        subprocess.Popen(f"python3 -m tdub.internal.cramped {rex_dir}", shell=True).wait()

    img_pairs = []
    for entry in mpl_dir.glob("*.png"):
        if "VRP" in entry.stem:
            pair = (mpl_dir / f"{entry.stem}.pdf", mpl_dir / f"{entry.stem}.png")
            img_pairs.append(pair)

    allrs_png_pair = ("matplotlib/allregions_pre.png", "matplotlib/allregions_post.png")
    r1j1b_png_pair = ("matplotlib/reg1j1b_preFit.png", "matplotlib/reg1j1b_postFit.png")
    r2j1b_png_pair = ("matplotlib/reg2j1b_preFit.png", "matplotlib/reg2j1b_postFit.png")
    r2j2b_png_pair = ("matplotlib/reg2j2b_preFit.png", "matplotlib/reg2j2b_postFit.png")
    allrs_pdf_pair = ("matplotlib/allregions_pre.pdf", "matplotlib/allregions_post.pdf")
    r1j1b_pdf_pair = ("matplotlib/reg1j1b_preFit.pdf", "matplotlib/reg1j1b_postFit.pdf")
    r2j1b_pdf_pair = ("matplotlib/reg2j1b_preFit.pdf", "matplotlib/reg2j1b_postFit.pdf")
    r2j2b_pdf_pair = ("matplotlib/reg2j2b_preFit.pdf", "matplotlib/reg2j2b_postFit.pdf")

    with doc:
        h1("Latest tW Fit")
        p("Generated {}".format(datetime.now().strftime("%b-%d-%Y %H:%M:%S")))
        with p():
            a("INT note link", href="https://cds.cern.ch/record/2667560")
        h2("Main Plots")
        p("Click images for PDF versions.")
        p("Prefit individuals")
        with div(cls="row"):
            with a(href=str(r1j1b_pdf_pair[0])):
                img(src=str(r1j1b_png_pair[0]), width=r"250px")
            with a(href=str(r2j1b_pdf_pair[0])):
                img(src=str(r2j1b_png_pair[0]), width=r"250px")
            with a(href=str(r2j2b_pdf_pair[0])):
                img(src=str(r2j2b_png_pair[0]), width=r"250px")
        p("Prefit combined")
        with div(cls="row"):
            with a(href=str(allrs_png_pair[0])):
                   img(src=str(allrs_png_pair[0]), cls="marginauto")
        p("Postfit individuals")
        with div(cls="row"):
            with a(href=str(r1j1b_pdf_pair[1])):
                img(src=str(r1j1b_png_pair[1]), width=r"250px")
            with a(href=str(r2j1b_pdf_pair[1])):
                img(src=str(r2j1b_png_pair[1]), width=r"250px")
            with a(href=str(r2j2b_pdf_pair[1])):
                img(src=str(r2j2b_png_pair[1]), width=r"250px")
        p("Postfit combined")
        with div(cls="row"):
            with a(href=str(allrs_png_pair[1])):
                   img(src=str(allrs_png_pair[1]), cls="marginauto")


        h2("Grouped Uncertainty Impacts (alphabetical then descending)")
        with div():
            p(raw(grouped_impacts_table(rex_dir, descending=False, tablefmt="html")))
            p(raw(grouped_impacts_table(rex_dir, descending=True, tablefmt="html")))

        h2("Other Plots 1j1b")
        with div() as d:
            add_images(
                d,
                ["pT_lep1", "pT_lep2", "pT_jet1", "met"],
                "1j1b",
                "pre",
                t="1j1b pre-fit kinematics",
            )
            add_images(
                d,
                ["pT_lep1", "pT_lep2", "pT_jet1", "met"],
                "1j1b",
                "post",
                t="1j1b post-fit kinematics",
            )
            add_images(
                d,
                ["pTsys_lep1lep2jet1met", "pT_jetS1", "cent_lep1lep2"],
                "1j1b",
                "pre",
                t="Top 3 1j1b BDT inputs pre-fit",
            )
            add_images(
                d,
                ["pTsys_lep1lep2jet1met", "pT_jetS1", "cent_lep1lep2"],
                "1j1b",
                "post",
                t="Top 3 1j1b BDT inputs post-fit",
            )

        h2("Other Plots 2j1b")
        with div() as d:
            add_images(
                d, ["pT_lep1", "pT_lep2"], "2j1b", "pre", t="2j1b pre-fit kinematics"
            )
            add_images(d, ["pT_jet1", "pT_jet2", "met"], "2j1b", "pre")

            add_images(
                d, ["pT_lep1", "pT_lep2"], "2j1b", "post", t="2j1b post-fit kinematics"
            )
            add_images(d, ["pT_jet1", "pT_jet2", "met"], "2j1b", "post")

            add_images(
                d,
                ["mass_lep1jet2", "mass_lep1jet1", "pTsys_lep1lep2jet1met"],
                "2j1b",
                "pre",
                t="Top 3 2j1b BDT inputs pre-fit",
            )
            add_images(
                d,
                ["mass_lep1jet2", "mass_lep1jet1", "pTsys_lep1lep2jet1met"],
                "2j1b",
                "post",
                t="Top 3 2j1b BDT inputs post-fit",
            )

        h2("other Plots 2j2b")
        with div() as d:
            add_images(
                d, ["pT_lep1", "pT_lep2"], "2j2b", "pre", t="2j2b pre-fit kinematics"
            )
            add_images(d, ["pT_jet1", "pT_jet2", "met"], "2j2b", "pre")
            add_images(
                d, ["pT_lep1", "pT_lep2"], "2j2b", "post", t="2j2b post-fit kinematics"
            )
            add_images(d, ["pT_jet1", "pT_jet2", "met"], "2j2b", "post")

            add_images(
                d,
                ["mass_lep1jet1", "mass_lep1jet2", "pT_jet2"],
                "2j2b",
                "pre",
                t="Top 3 2j2b BDT inputs pre-fit",
            )
            add_images(
                d,
                ["mass_lep1jet1", "mass_lep1jet2", "pT_jet2"],
                "2j2b",
                "post",
                t="Top 3 2j2b BDT inputs post-fit",
            )

    with open(Path(rex_dir) / "index.html", "w") as f:
        print(doc, file=f)



"""Cramped plot."""

import os
import pathlib
import sys

from typing import Tuple

import matplotlib.pyplot as plt

from tdub.art import canvas_from_counts, legend_last_to_first, draw_atlas_label
from tdub.rex import meta_text
from tdub.rex import region_plot_raw_material


def cramped(rex_dir, stage="pre"):
    """Generate a crampted plot."""
    fig: plt.Figure
    ax: Tuple[Tuple[plt.Axes, ...], ...]
    heights = [3.25, 1]
    fig, ax = plt.subplots(
        2,
        3,
        figsize=(11.5, 5.5),
        gridspec_kw=dict(
            width_ratios=[1, 1, 1],
            height_ratios=heights,
            hspace=0.15,
            wspace=0.020,
        ),
    )

    counts, errors, datagram, total_mc, uncertainty = region_plot_raw_material(
        rex_dir,
        "reg1j1b",
        stage,
        "tW",
    )
    bin_edges = datagram.edges
    canvas_from_counts(
        counts,
        errors,
        bin_edges,
        uncertainty=uncertainty,
        total_mc=total_mc,
        mpl_triplet=(fig, ax[0][0], ax[1][0]),
        combine_minor=True,
    )

    counts, errors, datagram, total_mc, uncertainty = region_plot_raw_material(
        rex_dir,
        "reg2j1b",
        stage,
        "tW",
    )
    bin_edges = datagram.edges
    canvas_from_counts(
        counts,
        errors,
        bin_edges,
        uncertainty=uncertainty,
        total_mc=total_mc,
        mpl_triplet=(fig, ax[0][1], ax[1][1]),
        combine_minor=True,
    )

    counts, errors, datagram, total_mc, uncertainty = region_plot_raw_material(
        rex_dir,
        "reg2j2b",
        stage,
        "tW",
    )
    bin_edges = datagram.edges
    canvas_from_counts(
        counts,
        errors,
        bin_edges,
        uncertainty=uncertainty,
        total_mc=total_mc,
        mpl_triplet=(fig, ax[0][2], ax[1][2]),
        combine_minor=True,
    )

    legend_last_to_first(ax[0][2], ncol=1, loc="upper right")
    draw_atlas_label(
        ax[0][0],
        follow_shift=0.280,
        extra_lines=[meta_text("reg1j1b", stage)],
        follow="",
    )

    y1, y2 = ax[0][1].get_ylim()
    y2 *= 0.7
    ax[0][0].set_ylim([y1, y2])
    ax[0][1].set_ylim([y1, y2])
    ax[0][2].set_ylim([y1, y2])

    ax[0][0].set_xticklabels([])
    ax[0][1].set_xticklabels([])
    ax[0][2].set_xticklabels([])

    ax[0][1].set_yticklabels([])
    ax[0][2].set_yticklabels([])
    ax[1][1].set_yticklabels([])
    ax[1][2].set_yticklabels([])

    ax[0][0].set_ylabel("Events", ha="right", y=1.0)
    ax[1][2].set_xlabel("BDT Response", ha="right", x=1.0)
    ax[1][0].set_ylabel("Data/MC")

    # ax[1][0].set_xticks([0.4, 0.5, 0.6, 0.7])
    ax[1][0].set_xticks([0.3, 0.4, 0.5, 0.6, 0.7])
    ax[1][1].set_xticks([0.2, 0.3, 0.4, 0.5, 0.6])
    # ax[1][1].set_xticks([0.25, 0.35, 0.45, 0.55, 0.65])
    ax[1][2].set_xticks([0.5, 0.6, 0.7])
    # ax[1][2].set_xticks([0.5, 0.55, 0.6, 0.65, 0.7])
    # ax[1][2].set_xticks([0.5, 0.55, 0.6, 0.65, 0.7, 0.75])

    ax[0][1].text(0.05, 0.925, "2j1b", transform=ax[0][1].transAxes, fontsize=14)
    ax[0][2].text(0.05, 0.925, "2j2b", transform=ax[0][2].transAxes, fontsize=14)

    ax[1][0].set_xlim([0.3, ax[1][0].get_xlim()[1]])
    ax[0][0].set_xlim([0.3, ax[1][0].get_xlim()[1]])

    ax[1][1].set_xlim([0.15, ax[1][1].get_xlim()[1]])
    ax[0][1].set_xlim([0.15, ax[1][1].get_xlim()[1]])

    fig.subplots_adjust(left=0.075)

    if not os.path.exists(rd / "matplotlib"):
        os.mkdir(rd / "matplotlib")

    fig.savefig(rd / "matplotlib" / f"allregions_{stage}.pdf")
    fig.savefig(rd / "matplotlib" / f"allregions_{stage}.png")


if __name__ == "__main__":
    rd = pathlib.Path(sys.argv[1])
    cramped(rd, stage="pre")
    cramped(rd, stage="post")



"""A module to aid working with histograms."""

from __future__ import annotations

# stdlib
import logging
from typing import Tuple, Optional, Union, Iterable

# ext
import numpy as np

log = logging.getLogger(__name__)


class SystematicComparison:
    """Systematic template histogram comparison.

    Attributes
    ----------
    nominal : numpy.ndarray
        Nominal histogram bin counts.
    up : numpy.ndarray
        Up variation histogram bin counts.
    down : numpy.ndarray
        Down variation histogram bin counts.
    percent_diff_up : numpy.ndarray
        Percent difference between nominal and up varation.
    percent_diff_down : numpy.ndaray
        Percent difference between nominald and down variation.

    """

    def __init__(
        self,
        nominal: np.ndarray,
        up: np.ndarray,
        down: np.ndarray,
    ) -> None:
        """Class constructor."""
        self.nominal = nominal
        self.up = up
        self.down = down
        self.percent_diff_up = (up - nominal) / nominal * 100.0
        self.percent_diff_down = (down - nominal) / nominal * 100.0

    @property
    def percent_diff_min(self) -> float:
        """float: minimum for percent difference."""
        return float(np.amin([self.percent_diff_up, self.percent_diff_down]))

    @property
    def percent_diff_max(self) -> float:
        """float: maximum for percent difference."""
        return float(np.amax([self.percent_diff_up, self.percent_diff_down]))

    @property
    def template_max(self) -> float:
        """float: maximum height of a variation."""
        return float(np.amax(np.concatenate([self.up, self.down])))

    @staticmethod
    def one_sided(nominal: np.ndarray, up: np.ndarray) -> SystematicComparison:
        """Generate components of a systematic comparion plot.

        Parameters
        ----------
        nominal : numpy.ndarray
            Histogram bin counts for the nominal template.
        up : numpy.ndarray
            Histogram bin counts for the "up" variation.

        Returns
        -------
        SystematicComparison
            The complete description of the comparison

        """
        diffs = nominal - up
        down = nominal + diffs
        return SystematicComparison(nominal, up, down)


def bin_centers(bin_edges: np.ndarray) -> np.ndarray:
    """Get bin centers given bin edges.

    Parameters
    ----------
    bin_edges : numpy.ndarray
       edges defining binning

    Returns
    -------
    numpy.ndarray
       the centers associated with the edges

    Examples
    --------
    >>> import numpy as np
    >>> from tdub.hist import bin_centers
    >>> bin_edges = np.linspace(25, 225, 11)
    >>> centers = bin_centers(bin_edges)
    >>> bin_edges
    array([ 25.,  45.,  65.,  85., 105., 125., 145., 165., 185., 205., 225.])
    >>> centers
    array([ 35.,  55.,  75.,  95., 115., 135., 155., 175., 195., 215.])

    """
    return (bin_edges[1:] + bin_edges[:-1]) * 0.5


def to_uniform_bins(bin_edges: np.ndarray):
    """Convert a set of variable width bins to arbitrary uniform bins.

    This will create a set of bin edges such that the bin centers are
    at whole numbers, i.e. 5 variable width bins will return an array
    from 0.5 to 5.5: [0.5, 1.5, 2.5, 3.5, 4.5, 5.5].

    Parameters
    ----------
    bin_edges : numpy.ndarray
        Array of bin edges.

    Returns
    -------
    numpy.ndarray
        The new set of uniform bins

    Examples
    --------
    >>> import numpy as np
    >>> from tdub.hist import to_uniform_bins
    >>> var_width = [0, 1, 3, 7, 15]
    >>> to_uniform_bins(var_width)
    array([0.5, 1.5, 2.5, 3.5, 4.5])

    """
    return np.arange(0.5, len(bin_edges) + 0.5, dtype=np.float64)


def edges_and_centers(
    bins: Union[int, Iterable], range: Optional[Tuple[float, float]] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Create arrays for edges and bin centers.

    Parameters
    ----------
    bins : int or sequence of scalers
       the number of bins or sequence representing bin edges
    range : tuple(float, float), optional
       the minimum and maximum defining the bin range (used if bins is integral)

    Returns
    -------
    :py:obj:`numpy.ndarray`
       the bin edges
    :py:obj:`numpy.ndarray`
       the bin centers

    Examples
    --------
    from bin multiplicity and a range

    >>> from tdub.hist import edges_and_centers
    >>> edges, centers = edges_and_centers(bins=20, range=(25, 225))

    from pre-existing edges

    >>> edges, centers = edges_and_centers(np.linspace(0, 10, 21))

    """
    if isinstance(bins, int):
        if range is None:
            raise ValueError("for integral bins we require the range argument")
        edges = np.linspace(range[0], range[1], bins + 1)
    else:
        edges = np.asarray(bins)
        if not np.all(edges[1:] >= edges[:-1]):
            raise ValueError("bins edges must monotonically increase")
    centers = bin_centers(edges)
    return edges, centers
